package org.jutils.log;
/*------------------------------------------------------------------------------
Name:      LogManager.java
Project:   jutils.org
Copyright: jutils.org, see jutils-LICENSE file
Comment:   This logging manager object handles multiple 'named' logging
           instances (channels) which could be generated by one program for
           different logging purposes or which could be generated by many
           programs in the same VM. Each name must be unique.
------------------------------------------------------------------------------*/

import java.util.Hashtable;

import org.jutils.JUtilsException;
import org.jutils.runtime.StackTrace;


/*******************************************************************************
This LogManager does allow you to store and access multiple LogChannel objects
via names over this global object.
<p />
It also provides a default LogChannel named 'default' (you should
use the static String field LogManager.DEFAULT_CHANNEL to refer to it) which
publishes its output on the console, but this channel should be used only
as fallback or to do logging before having initialized your own LogChannel
objects.
<p />
The following example shows you how to use the LogChannel object and how
to register this object at the LogManager:
<pre>
   LogChannel tmp = new LogChannel(null);
   tmp.addLogLevel(LOG_ALL);                        // set loglevel
   tmp.addLogDevice(new LogDeviceConsole(tmp));     // create console
   LogManager.addLogChannel(DEFAULT_CHANNEL, tmp);  // register logchannel
</pre>

@author  Juergen Freidling
*******************************************************************************/
public class LogManager
   implements LogConstants
{
   private final static String DEFAULT_CHANNEL = "default";
   private static Hashtable channels = new Hashtable();
   // static pre initialization
   static {
      initialize();
   };




   /****************************************************************************
    * This method does the basic initialization of this object with the
    * default LogChannel object. <p />
    * This method is called automatically at class initialization, so you
    * should only use it if you want to re-initialize the default LogChannel
    * object to the default settings.
    ***************************************************************************/
   public static void initialize() {
      LogChannel tmp = new LogChannel(null);
      tmp.addLogLevel(LOG_ALL);
      tmp.addLogDevice(new LogDeviceConsole(tmp));
      LogManager.addLogChannel(DEFAULT_CHANNEL, tmp);
   }


   /****************************************************************************
    * With this method you can add a self created LogChannel object to this
    * manager object.
    *
    * @param      channelName
    *             The name under which this LogChannel object should be stored.
    * @param      log
    *             The LogChannel object.
    ***************************************************************************/
   public static boolean addLogChannel(String channelName, LogChannel log) {
      if ((channelName != null) && (!channelName.equals("")) && (log != null)) {
         channels.put(channelName, log);
         return true;
      }
      return false;
   }


   /****************************************************************************
    * This method gives you direct access to the LogChannel object with the
    * specified name.
    *
    * @param      channelName
    *             The name of the LogChannel object which should be returned.
    * @return     the LogChannel object with the given name or null if no
    *             object with this name does exist.
    ***************************************************************************/
   public static LogChannel getLogChannel(String channelName) {
      LogChannel log = (LogChannel) channels.get(channelName);
      return log;
   }


   /****************************************************************************
    * This is the basic logging method which is also used by all other
    * logging methods like error() or warn() which defines the loglevel
    * itself.
    *
    * @param      channelName
    *             The name of the LogChannel object to which the logging
    *             information should be sent.
    * @param      level
    *             The loglevel value. Only the values which are defined in
    *             LogConstants should be used.
    * @param      source
    *             An identifier name from whom this message was sent.
    * @param      text
    *             The data which should be logged.
    ***************************************************************************/
   public static void log(String channelName,
                          int level,
                          String source,
                          String text)
   {
      if ((channelName == null) || (channelName.equals(""))) {
         log(DEFAULT_CHANNEL, level, source, text);
         return;
      }
      LogChannel channel = (LogChannel) channels.get(channelName);
      if (channel != null) {
         channel.log(level, source, text);
      }
      else {
         if (!channelName.equals(DEFAULT_CHANNEL)) {
            log(DEFAULT_CHANNEL, LOG_ERROR, source,
                "LogManager.java - logchannel '" + channelName + "' not found! [" + text + "]");
         }
      }
   }


   /****************************************************************************
    * A easy to use logging method which does log to the default channel in
    * plain text style an with no source identifier.
    *
    * @param      text
    *             The data which should be logged.
    ***************************************************************************/
   public static void log(String text) {
      log(DEFAULT_CHANNEL, LOG_PLAIN, "", text);
   }


   /****************************************************************************
    * Use this for reporting and tracing exceptions. The data will be
    * published as error data.
    *
    * @param      channelName
    *             the selection, to which channel this log should be sent
    * @param      source
    *             the identifier of the sender
    * @param      t
    *             the throwable which to trace.
    ***************************************************************************/
   public static void exception(String channelName,
                                String source,
                                Throwable t) {
      String text = StackTrace.getStackTrace(t);
      log(channelName, LOG_ERROR, source, text);
   }


   /****************************************************************************
    * Use this for reporting internal errors. Depending on the
    * system / application / situation the execution must be stopped.
    * This stop must be done in the application.
    *
    * @param      channelName
    *             the selection, to which channel this log should be sent
    * @param      source
    *             the identifier of the sender
    * @param      text
    *             the string to log
    ***************************************************************************/
   public static void error(String channelName, String source, String text) {
      log(channelName, LOG_ERROR, source, text);
   }


   /****************************************************************************
    * Use this for logging output which is important, for example a
    * 'login denied' event.
    *
    * @param      channelName
    *             the selection, to which channel this log should be sent
    * @param      source
    *             the identifier of the sender
    * @param      text
    *             the string to log
    ***************************************************************************/
   public static void warn(String channelName, String source, String text) {
      log(channelName, LOG_WARN, source, text);
   }


   /****************************************************************************
    * Use this for normal logging output.
    *
    * @param      channelName
    *             the selection, to which channel this log should be sent
    * @param      source
    *             the identifier of the sender
    * @param      text
    *             the string to log
    ***************************************************************************/
   public static void info(String channelName, String source, String text) {
      log(channelName, LOG_INFO, source, text);
   }


   /****************************************************************************
    * Tracing when entering methods.
    *
    * @param      channelName
    *             the selection, to which channel this log should be sent
    * @param      source
    *             the identifier of the sender
    * @param      text
    *             the string to log
    ***************************************************************************/
   public static void call(String channelName, String source, String text) {
      log(channelName, LOG_CALL, source, text);
   }


   /****************************************************************************
    * Output of performant measurements (elapsed milliseconds).
    *
    * @param      channelName
    *             the selection, to which channel this log should be sent
    * @param      source
    *             the identifier of the sender
    * @param      text
    *             the string to log
    ***************************************************************************/
   public static void time(String channelName, String source, String text) {
      log(channelName, LOG_TIME, source, text);
   }


   /****************************************************************************
    * Trace the execution.
    *
    * @param      channelName
    *             the selection, to which channel this log should be sent
    * @param      source
    *             the identifier of the sender
    * @param      text
    *             the string to log
    ***************************************************************************/
   public static void trace(String channelName, String source, String text) {
      log(channelName, LOG_TRACE, source, text);
   }


   /****************************************************************************
    * Log full state information of an object.
    *
    * @param      channelName
    *             the selection, to which channel this log should be sent
    * @param      source
    *             the identifier of the sender
    * @param      text
    *             the string to log
    ***************************************************************************/
   public static void dump(String channelName, String source, String text) {
      log(channelName, LOG_DUMP, source, text);
   }


   /****************************************************************************
    * Log text as given, no extra data attached.
    *
    * @param      channelName
    *             the selection, to which channel this log should be sent
    * @param      source
    *             the identifier of the sender
    * @param      text
    *             the string to log
    ***************************************************************************/
   public static void plain(String channelName, String source, String text) {
      log(channelName, LOG_PLAIN, source, text);
   }


   /****************************************************************************
    * Testing facility
    ***************************************************************************/
   public static void main(String[] args) {
      // Test 1
      System.out.println("\n### BEGIN TEST 1 ###");
      LogManager.log("Test1");
      LogManager.log("default", LOG_PLAIN, "LogManager", "Test1");
      LogManager.log("default", LOG_ERROR, "LogManager", "Test1");
      System.out.println("### END TEST 1 ###\n");


      // Test 2
      System.out.println("\n### BEGIN TEST 2 ###");
      LogChannel log1 = new LogChannel(null);
      LogDeviceConsole console1 = new LogDeviceConsole(log1);
      console1.setColorMode(true);
      log1.addLogDevice(console1);
      LogManager.addLogChannel("log1", log1);
      LogChannel log2 = new LogChannel(null);
      LogDeviceConsole console2 = new LogDeviceConsole(log2);
      console2.setColorMode(false);
      log2.addLogDevice(console2);
      LogManager.addLogChannel("log2", log2);
      LogManager.log("log1", LOG_INFO, "LogManager.main()", "Output1");
      LogManager.log("log2", LOG_ERROR, "LogManager.main()", "Output2");
      System.out.println("### END TEST 2 ###\n");


      // Test 3
      System.out.println("\n### BEGIN TEST 3 ###");
      LogManager.error("log1", "LogManager.main()", "3_1");
      LogManager.warn("log1", "LogManager.main()", "3_2");
      LogManager.info("log1", "LogManager.main()", "3_3");
      LogManager.call("log1", "LogManager.main()", "3_4");
      LogManager.time("log1", "LogManager.main()", "3_5");
      LogManager.trace("log1", "LogManager.main()", "3_6");
      LogManager.plain("log1", "LogManager.main()", "3_7");
      System.out.println("### END TEST 3 ###\n");

   }
}
