package org.jutils.pool;

/*------------------------------------------------------------------------------
Name:      PoolManager.java
Project:   jutils.org
Copyright: jutils.org, see jutils-LICENSE file
Comment:   Basic handling of a pool of limited resources
Version:   $Id: PoolManager.java,v 1.8 2000/07/13 08:43:16 winkler Exp $
         $Source: /opt/cvsroot/jutils/src/java/org/jutils/pool/PoolManager.java,v $
Author:    ruff@swand.lake.de
------------------------------------------------------------------------------*/
import org.jutils.JUtilsException;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;


/**
 * A little framework to handle a pool of limited resources.
 * <p />
 * You can use this pool as a base class handling of your limited resources
 * like a 'JDBC connection pool' or a 'thread pool'.
 * <p />
 * The important attributes of a resource are gathered in the <code>ResourceWrapper</code> class.
 * <p />
 * You can easily handle bigger number of resources with good performance.
 * <p />
 * To find out how to use it see the TestPool example in the main() method of this class.
 * <p />
 * <pre>
 *    State chart of resource handling:
 *
 *      +<------- reserve() if(numIdle==0) ---------------------+
 *      |                                                       |
 *      |     +<- reserve() --+        +<-preReserve() *-+      |
 *      |     | if(numIdle>0) |        |                 |      |
 *      |     |               |        |                 |      |
 *    #########               ##########                 ##########
 *   #         #             #          #               #          #
 *   #  busy   #             #  idle    #               #  undef   #
 *   #         #             #          #               #          #
 *    #########               ##########                 ##########
 *      |  |  |               | |  | | |                 | |  |  |
 *      |  |  |   Explicit    | |  | | |   Explicit      | |  |  |
 *      |  |  +-- release() ->+ |  | | +-- erase() ----->+ |  |  |
 *      |  |                    |  | |                     |  |  |
 *      |  +busyToIdleTimeout()>+  | +idleToEraseTimeout()>+  |  |
 *      |                          |                          |  |
 *      |                          +-- erase on max cycles *->+  |
 *      |                                                        |
 *      +--------- busyToEraseTimeout() since creation *-------->+
 *
 * </pre>
 * There are three states:
 * <ul>
 *    <li>busy: The resource is in use
 *    </li>
 *    <li>idle: The resource is allocated and ready but not in use
 *    </li>
 *    <li>undef: The resource is not yet allocated or it is erased again
 *    </li>
 * </ul>
 * <p>
 * Note that state transitions marked with '*' are not yet implemented.
 * If you need one of them, code it and contribute it please.
 * <p>
 * You can choose which states you wish for your resource and how timeouts
 * are used to handle state transitions.<br />
 * </p>
 * <p>
 * For example if you want to pool user login sessions and want to do
 * an auto logout after 60 minutes, you would use the busyToIdleTimeout and set it to 60*60*1000.<br />
 * If a user is active you can refresh the session with busyRefresh().<br />
 * Often you want to use your own generated sessionId as the primary key
 * for this resource, you can pass it as the instanceId argument to reserve(sessionId).<br />
 * (See example [2] in this main() method)
 * </p>
 * <p>
 * If you want to pool JDBC connections, reserve() a connection before you do your query
 * and release() it immediately again. If you
 * want to close connections after peak usage, you could set a idleToEraseTimeout,
 * to erase your JDBC connection after some time not used (reducing the current pool size).<br />
 * Note that in this example the connections are anonymous (all are logged in to the database
 * with the same user name), it is not important which you receive.<br />
 * (See example [1] in this main() method)
 * </p>
 *
 *
 * @author "Marcel Ruff" <ruff@swand.lake.de>
 * @see org.jutils.pool.ResourceWrapper
 */
public class PoolManager
{
   /** Nice, unique name for logging output */
   private String ME = "PoolManager";
   /** A nice name for the generated id */
   private String poolName = "Resource";
   /** The callback into the using application, which needs to implement the interface I_PoolManager */
   private I_PoolManager callback = null;
   /** Holds busy resources */
   private java.util.Hashtable busy = null;
   /** Holds free resources */
   private java.util.Vector idle = null;
   /** Default maximum pool size (number of resources) */
   private int maxInstances = 100;
   /** Default maximum busy time of a resource, on timeout it changes state from 'busy' to 'idle' */
   private long busyToIdleTimeout = 0;
   /* Default maximum life span of a resource, on timeout it changes state from 'busy' to 'undef' (it is deleted) */
   //private long busyToEraseTimeout = 0;
   /** Default maximum idle span of a resource, on timeout it changes state from 'idle' to 'undef' (it is deleted) */
   private long idleToEraseTimeout = 0;
   /** Unique counter to generate IDs */
   private long counter = 1;
   /** Use this constant with the reserve() method, the PoolManager uses the hashCode() of your object */
   public final static String USE_HASH_CODE = "h";
   /** Use this constant with the reserve() method, the PoolManager uses the toString() of your object, if no toString() exists the object reference is used, e.g. "oracle.jdbc.driver.OracleConnection@443226" */
   public final static String USE_OBJECT_REF = "o";
   /** Use this constant with the reserve() method, the PoolManager generates a random, unique in universe session ID */
   public final static String GENERATE_RANDOM = "r";

   private final Object meetingPoint = new Object();

   /**
   * Create a new pool instance with the desired behavior.
   * <p />
   * Implementation:<br />
   * There are two list in this class
   * <ul>
   *    <li>busy: Holds resources which are currently busy</li>
   *    <li>idle: Holds resources in the free pool, waiting to be reused</li>
   * </ul>
   * If you ask for a new resource, this will move into the busy list<br />
   * Is a resource released or a specified timeout occurred, it is moved into the idle list.
   * <p />
   * @param poolName       A nice name for this pool manager instance.
   * @param callback       The interface 'I_PoolManager' callback
   * @param maxInstances   Max. number of resources in this pool.
   * @param busyToIdleTimeout Max. busy time of this resource in milli seconds<br />
   *                       On timeout it changes state from 'busy' to 'idle'.<br />
   *                       You can overwrite this value for each resource instance<br />
   *                       0 switches it off<br />
   *                       You get called back through I_PoolManager.busyToIdle() on timeout
   *                       allowing you to code some specific handling.
   * @param idleToEraseTimeout Max. idle time span of this resource in milli seconds<br />
   *                     On timeout it changes state from 'idle' to 'undef' (it is deleted).<br />
   *                     You can overwrite this value for each resource instance<br />
   *                     0 switches it off<br />
   *                     You get called back through I_PoolManager.toErased() on timeout
   *                     allowing you to code some specific handling.
   */
   public PoolManager(String poolName, I_PoolManager callback, int maxInstances, long busyToIdleTimeout, long idleToEraseTimeout) {
      this.poolName = poolName; // e.g. "SessionId"
      this.callback = callback;
      this.ME = poolName + "-PoolManager"; // e.g. "SessionId-PoolManager"
      this.setMaxInstances(maxInstances);
      this.setBusyToIdleTimeout(busyToIdleTimeout);
      this.setIdleToEraseTimeout(idleToEraseTimeout);
      this.busy = new java.util.Hashtable(maxInstances);
      this.idle = new java.util.Vector(maxInstances);
      /*
        if (busyToIdleTimeout > 0 || idleToEraseTimeout > 0) {
       System.err.println(ME + ": Created PoolManager with a maximum of " + maxInstances + " instances" +
        " and a default busy recycling-timeout of " + TimeHelper.millisToNice(this.busyToIdleTimeout) +
        " and a default idle recycling-timeout of " + TimeHelper.millisToNice(this.idleToEraseTimeout));
        }
        else
       System.err.println(ME + ": Created PoolManager with a maximum of " + maxInstances + " instances without auto-recycling.");
        */
   }

   /**
   * Set the maximum pool size.
   * @param maxInstances How many resources are allowed
   */
   private void setMaxInstances(int maxInstances) {
      this.maxInstances = maxInstances;
   }

   /**
   * On timeout, the resource changes state from 'busy' to 'idle'.<br />
   * @param busyToIdleTimeout Max. busy time of this resource in milli seconds<br />
   *                       On timeout it changes state from 'busy' to 'idle'.<br />
   *                       You can overwrite this value for each resource instance<br />
   *                       0 switches it off
   */
   private void setBusyToIdleTimeout(long busyToIdleTimeout) {
      if (busyToIdleTimeout > 0 && busyToIdleTimeout < 100) {
         // System.err.println(ME + ": Setting minimum release timeout from " + busyToIdleTimeout + " to 100 milli seconds.");
         this.busyToIdleTimeout = 100;
      }
      else
         this.busyToIdleTimeout = busyToIdleTimeout;
   }

   /**
   * Set the max. life span of the resources.
   * @param idleToEraseTimeout Max. idle time of this resource in milli seconds<br />
   *                       You can overwrite this value for each resource instance<br />
   *                       0 switches it off
   */
   private void setIdleToEraseTimeout(long idleToEraseTimeout) {
      if (idleToEraseTimeout > 0 && idleToEraseTimeout < 100) {
         // System.err.println(ME + ": Setting minimum idleErase timeout from " + idleToEraseTimeout + " to 100 milli seconds.");
         this.idleToEraseTimeout = 100;
      }
      else
         this.idleToEraseTimeout = idleToEraseTimeout;
   }

   /**
   * Get a new resource.
   * <p />
   * The life span is set to the default value of the pool.
   * <p />
   * The instance Id is random and unique generated.
   * @exception JUtilsException Error with random generator
   */
   public ResourceWrapper reserve() throws JUtilsException {
      return reserve(busyToIdleTimeout, idleToEraseTimeout, GENERATE_RANDOM);
   }

   /**
   * Get a new resource.
   * <p />
   * The life span is set to the default value of the pool.
   * <p />
   * @parameter instanceId See description in other reserve() method.
   * @exception JUtilsException Error with random generator
   */
   public ResourceWrapper reserve(String instanceId) throws JUtilsException {
      return reserve(busyToIdleTimeout, idleToEraseTimeout, instanceId);
   }

   /**
   * Get a new resource.
   *
   * @param localBusyToIdleTimeout Max. busy time of this resource in milli seconds,
   *                       only for this current resource<br />
   *                       Overwrite locally the default<br />
   *                       0 switches busy timeout off
   *
   * @param localIdleToEraseTimeout Max. idle time of this resource in milli seconds,
   *                       only for this current resource<br />
   *                       Overwrite locally the default<br />
   *                       0 switches idle timeout off
   *
   * @param instanceId If given and string length > 1, the delivered ID is used:
   *             If in busy list found, this is returned, else a new is created.<br />
   *             USE_HASH_CODE   - The PoolManager generates a simple one (hashCode())<br />
   *             USE_OBJECT_REF  - The object reference is used<br />
   *             GENERATE_RANDOM - The PoolManager generates a random, unique in universe session ID
   *
   * @return rw The resource handle (always != null)
   *
   * @exception JUtilsException Error with random generator
   */
   public ResourceWrapper reserve(long localBusyToIdleTimeout, long localIdleToEraseTimeout, String instanceId) throws JUtilsException {
      synchronized (meetingPoint) {
         ResourceWrapper rw = null;
         if (instanceId != null && instanceId.length() > 1) {
            rw = findBusySilent(instanceId);
            if (rw != null) {
               // System.out.println(ME + ": Reconnected to busy resource '" + instanceId + "' ...");
               return rw;
            }
         }
         if (idle.size() > 0) {
            // System.out.println(ME + ": Resource is recycled from idle pool ...");
            rw = (ResourceWrapper) idle.lastElement();
            rw.init(createId(instanceId), rw.getResource(), localBusyToIdleTimeout, localIdleToEraseTimeout);
            swap(rw, true);
            // System.out.println(ME + ": Access on cached resource '" + rw.getInstanceId() + "' granted");
            // dumpState(rw.getInstanceId());
         }
         else
            if (busy.size() >= maxInstances) {
               throw new JUtilsException("ResourceExhaust", "Sorry, " + maxInstances + " resources consumed, no more resources available");
            }
            else {
               instanceId = createId(instanceId);
               Object resource = callback.toCreate(instanceId);
               rw = new ResourceWrapper(this, instanceId, resource, localBusyToIdleTimeout, idleToEraseTimeout);
               rw.toBusy();
               busy.put(rw.getInstanceId(), rw);
               // System.out.println(ME + ": Granted access to new resource '" + rw.getInstanceId() + "'");
               // dumpState(rw.getInstanceId());
            }
         return rw;
      }
   }

   /**
   * Release a resource explicitly from 'busy' into the 'idle' pool.
   * @param instanceId The unique resource ID
   * @exception JUtilsException
   */
   public void release(String instanceId) throws JUtilsException {
      synchronized (meetingPoint) {
         swap(findLow(instanceId), false);
      }
      //System.out.println(ME + ": Resource <" + instanceId + "> released explicitly");
      // dumpState(instanceId);
   }

   /**
   *  Generate a unique resource ID <br>
   *
   *  @instanceId USE_HASH_CODE=null   ResourceWrapper generates a simple ID = resource.hashCode()<br />
   *              "..."                Your supplied ID is used<br />
   *              GENERATE_RANDOM=""   A random ID is generated, unique in universe and over time
   *  @return unique ID or null
   *  @exception JUtilsException random generator
   */
   private String createId(String instanceId) throws JUtilsException {
      if (instanceId == null || instanceId.equals(GENERATE_RANDOM) == false)
         return instanceId; // ResourceWrapper generates a simple id

      try {
         String ip;
         try {
            java.net.InetAddress addr = java.net.InetAddress.getLocalHost();
            ip = addr.getHostAddress();
         }
         catch (Exception e) {
            // System.err.println(ME + ": Can't determin your IP address");
            ip = "localhost";
         }

         // This is a real random, but probably not necessary here:
         // Random random = new java.security.SecureRandom();
         java.util.Random ran = new java.util.Random(); // is more or less currentTimeMillis

         // Note: We should include the process ID from this JVM on this host to be granted unique

         //   <IP-Address>-<InstanceName>-<TimestampMilliSec>-<RandomNumber>-<LocalCounter>
         return ip + "-" + poolName + "-" + System.currentTimeMillis() + "-" + ran.nextInt() + "-" + (counter++);
      }
      catch (Exception e) {
         String text = "Can't generate a unique instanceId: " + e.toString();
         throw new JUtilsException("ResourceNoId", text);
      }
   }

   /**
   * Idle - busy swapper.
   */
   private void swap(ResourceWrapper rw, boolean toBusy) {
      if (rw == null) {
         System.err.println("##########INTERNAL PROBLEM IN PollManager");
         return;
      }
      if (toBusy) {
         idle.removeElementAt(idle.size() - 1);
         busy.put(rw.getInstanceId(), rw);
         rw.toBusy();
         callback.idleToBusy(rw.getResource());
         //System.err.println("Swapped to busy:" + rw.getInstanceId());
      }
      else { // recycling ...
         busy.remove(rw.getInstanceId());
         idle.addElement(rw);
         rw.toIdle();
         callback.busyToIdle(rw.getResource());
         //System.err.println("Swapped to idle:" + rw.getInstanceId());
      }
   }

   /**
   * Find a resource in busy list.
   * @param instanceId The unique resource ID
   * @return The handle containing the resource.
   */
   private ResourceWrapper findBusySilent(String instanceId) {
      if (instanceId == null)
         return null;
      return (ResourceWrapper) busy.get(instanceId);
   }

   /**
   * Find a resource in idle list.
   * <p />
   * Note that this is currently a linear search (not high performing).
   * @param instanceId The unique resource ID
   * @return The handle containing the resource.
   */
   private ResourceWrapper findIdleSilent(String instanceId) {
      if (instanceId == null)
         return null;
      for (int ii = 0; ii < idle.size(); ii++) {
         ResourceWrapper rw = (ResourceWrapper) idle.elementAt(ii);
         if (rw.getInstanceId().equals(instanceId))
            return rw;
      }
      return null;
   }

   /**
   * Find a resource in busy list.
   * @param instanceId The unique resource ID
   * @return Handle containing resource
   * @exception JUtilsException "ResourceNotFound"
   */
   private ResourceWrapper findLow(String instanceId) throws JUtilsException {
      if (instanceId == null) {
         String text = "Your resource ID is null";
         throw new JUtilsException("ResourceNotFound", text);
      }
      ResourceWrapper rw = findBusySilent(instanceId);
      if (rw == null) {
         String text = "Resource '" + instanceId + "' is invalid, timed out?";
         throw new JUtilsException("ResourceNotFound", text);
      }
      return rw;
   }

   /**
   * Test if the resource is busy.
   *
   * @param instanceId The unique resource ID
   * @return true - is in 'busy' state<br />
   *         false - is in 'idle' state or unknown
   */
   public boolean isBusy(String instanceId) {
      ResourceWrapper rw = findBusySilent(instanceId);
      return (rw != null);
   }

   /**
   * Restart countdown for resource life cycle.
   * <p />
   * Rewind the timeout for 'busy' to 'idle' transition.
   * @param instanceId The unique resource ID
   * @exception JUtilsException ResourceNotFound
   */
   public void busyRefresh(String instanceId) throws JUtilsException {
      synchronized (meetingPoint) {
         ResourceWrapper rw = findLow(instanceId);
         rw.touchBusy();
      }
   }

   /**
   * Number of resources in the 'busy' list.
   * @return Number of 'busy' resources
   */
   public int getNumBusy() {
      return busy.size();
   }

   /**
   * Number of resources in the 'idle' list.
   * @return Number of 'idle' resources
   */
   public int getNumIdle() {
      return idle.size();
   }

   /**
   * Cleanup.
   */
   protected void finalize() throws Throwable {
      destroy();
   }

   /**
   * Dump the current state of this pool.
   */
   public String getState() {
      return "Overview for '" + ME + "':" + " BUSY=" + busy.size() + " IDLE=" + idle.size() + " MAX_AVAILABLE=" + (maxInstances - busy.size());
   }

   /**
   * Dump state of this object into a XML ASCII string.
   * <p />
   * @return internal state of this PoolManager as a XML ASCII string
   */
   public final String toXml() {
      return toXml((String) null);
   }

   /**
   * Dump state of this object into a XML ASCII string.
   * <p />
   * @param extraOffset indenting of tags for nice output
   * @return internal state of this PoolManager as a XML ASCII string
   */
   public final String toXml(String extraOffset) {
      StringBuffer buf = new StringBuffer();
      String offset = "\n";
      if (extraOffset == null)
         extraOffset = "";
      offset += extraOffset;
      synchronized (meetingPoint) {
         buf.append(offset).append("<").append(ME).append(" maxInstances='").append(maxInstances);
         buf.append("' busyToIdleTimeout='").append(busyToIdleTimeout);
         buf.append("' idleToEraseTimeout='").append(idleToEraseTimeout).append("'>");
         buf.append(offset).append("   <busy num='").append(busy.size()).append("'>");
         for (Enumeration e = busy.elements(); e.hasMoreElements();) {
            ResourceWrapper rw = (ResourceWrapper) e.nextElement();
            buf.append(rw.toXml(extraOffset + "   "));
         }
         buf.append(offset).append("   </busy>");
         buf.append(offset).append("   <idle num='").append(idle.size()).append("'>");
         for (int ii = 0; ii < idle.size(); ii++) {
            ResourceWrapper rw = (ResourceWrapper) idle.elementAt(ii);
            buf.append(rw.toXml(extraOffset + "   "));
         }
         buf.append(offset).append("   </idle>");
         buf.append(offset).append("</" + ME + ">");
      }
      return buf.toString();
   }

   /**
   * Recycle busy resource after timeout.
   * <p />
   * This method is a callback from ResourceWrapper
   * @parameter userData The ResourceWrapper object
   */
   void timeoutBusyToIdle(ResourceWrapper rw) {
      // System.out.println(ME + ": Resource '" + rw.getInstanceId() + "' is recycled from 'busy' state, was not in use since " + TimeHelper.millisToNice(rw.busyElapsed()));
      synchronized (meetingPoint) {
         swap(rw, false);
      }
      // dumpState(rw.getInstanceId());
   }

   /**
   * Erase an idle resource after timeout.
   * <p />
   * This method is a callback from ResourceWrapper
   * @parameter userData The ResourceWrapper object
   */
   void timeoutIdleToErase(ResourceWrapper rw) {
      // System.out.println(ME + ": Resource '" + rw.getInstanceId() + "' is erased from 'idle' state after timeout");
      erase(rw);
   }

   /**
   * Explicitly remove a resource.
   * <p />
   * It is erased if it is found in the idle list or even when it is busy.
   * @param instanceId The unique resource ID
   */
   public void erase(String instanceId) {
      if (instanceId == null || instanceId.length() < 1)
         return;
      synchronized (meetingPoint) {
         ResourceWrapper rw = findBusySilent(instanceId);
         if (rw == null) {
            rw = findIdleSilent(instanceId);
         }
         erase(rw);
      }
   }

   /**
   * Remove a resource.
   * <p />
   * Remove it from the 'idle' or the 'busy' list.
   * @param rw The resource wrapper object
   */
   private void erase(ResourceWrapper rw) {
      if (rw == null)
         return;
      synchronized (meetingPoint) {
         if (busy.get(rw.getInstanceId()) != null) {
            swap(rw, false); // move to idle ... calls busyToIdle() callback
            //busy.remove(rw.getInstanceId());
         }
         idle.remove(rw);
         callback.toErased(rw.getResource());
         rw.destroy();
      }
      // dumpState(rw.getInstanceId());
   }

   /**
   * Cleanup everything.
   */
   public void destroy() {
      // System.out.println(ME + ": Freeing all resources, " + busy.size() + " are busy and " + idle.size() + " idle.");
      synchronized (meetingPoint) {
         java.util.Enumeration e = busy.elements();
         for (; e.hasMoreElements();) {
            erase((ResourceWrapper) e.nextElement());
         }
         busy.clear();
         for (int ii = 0; ii < idle.size(); ii++) {
            ResourceWrapper rw = (ResourceWrapper) idle.elementAt(ii);
            callback.toErased(rw.getResource());
            rw.destroy();
         }
         idle.removeAllElements();
      }
   }

   /**
   * For testing only.
   * <p />
   * Invoke: java org.jutils.pool.PoolManager
   */
   public static void main(String[] args) {
      final String ME = "TestPool";

      /**
         * This class is usually a UserSession object or a JDBC connection object
         * or whatever resource you want to handle
         */
      class TestResource {
         String name;
         String instanceId;
         boolean isBusy;
         boolean isErased = false;
         public TestResource(String name, String instanceId, boolean isBusy) {
            this.name = name;
            this.instanceId = instanceId;
            this.isBusy = isBusy;
         }
         public String toString() {
            return name;
         }
      }

      /**
         * This class does the resource pooling for TestResource,
         * with the help of PoolManager
         */
      class TestPool implements I_PoolManager {
         private int counter = 0;
         PoolManager poolManager;
         TestPool(int maxInstances, long busyToIdle, long idleToErase) {
            poolManager = new PoolManager(ME, this, maxInstances, busyToIdle, idleToErase);
         }

         // These four methods are callbacks from PoolManager ...
         public void idleToBusy(Object resource) {
            TestResource rr = (TestResource) resource;
            System.out.println("Entering idleToBusy(" + rr.name + ") ...");
            rr.isBusy = true; // you could do some re-initialization here ...
         }
         public void busyToIdle(Object resource) {
            TestResource rr = (TestResource) resource;
            System.out.println("Entering busyToIdle(" + rr.name + ") ...");
            rr.isBusy = false; // you could do some coding here ...
         }
         public Object toCreate(String instanceId) throws JUtilsException {
            TestResource rr = new TestResource("TestResource-" + (counter++), instanceId, true);
            System.out.println("Entering toCreate(instanceId='" + instanceId + "', " + rr.name + ") ...");
            return rr;
         }
         public void toErased(Object resource) {
            TestResource rr = (TestResource) resource;
            System.out.println("Entering toErased(" + rr.name + ") ...");
            rr.isErased = true;
         }

         // These methods are used by your application to get a recource ...
         TestResource reserve() {
            return reserve(GENERATE_RANDOM);
         }
         TestResource reserve(String instanceId) {
            System.out.println("Entering reserve(" + instanceId + ") ...");
            try {
               synchronized (this) {
                  ResourceWrapper rw = (ResourceWrapper) poolManager.reserve(instanceId);
                  TestResource rr = (TestResource) rw.getResource();
                  if (rr == null)
                     PoolManager.panic("*****ERROR:rr==null");
                  rr.instanceId = rw.getInstanceId(); // remember the generated unique id
                  return rr;
               }
            }
            catch (JUtilsException e) {
               System.err.println("*****WARNING:Caught exception in reserve(): " + e.reason);
               return null;
            }
         }
         void release(TestResource rr) {
            System.out.println("Entering release() ...");
            try {
               synchronized (this) {
                  poolManager.release(rr.instanceId);
               }
            }
            catch (JUtilsException e) {
               System.err.println("*****WARNING:Caught exception in release(): " + e.reason);
            }
         }
      }

      // And now test it ...

      // [1] Test with generated instance ID and busy timeout
      if (true) {
         System.out.println("\n\n=========================\nStarting TEST 1 ...");
         TestPool testPool = new TestPool(3, 2000, 0);
         TestResource r0 = testPool.reserve();
         TestResource r1 = testPool.reserve(GENERATE_RANDOM); // is default
         TestResource r2 = testPool.reserve();
         testPool.reserve();
         if (testPool.poolManager.getNumBusy() != 3 || testPool.poolManager.getNumIdle() != 0)
            PoolManager.panic("TEST 1.1 FAILED: Wrong number of busy/idle resources");
         testPool.release(r0);
         if (testPool.poolManager.getNumBusy() != 2 || testPool.poolManager.getNumIdle() != 1)
            PoolManager.panic("TEST 1.2 FAILED: Wrong number of busy/idle resources");
         testPool.reserve();
         testPool.release(r2);
         System.out.println(testPool.poolManager.toXml());

         // The resources are swapped to idle in 2 seconds, lets wait 3 seconds ...
         try {
            Thread.currentThread().sleep(3000);
         }
         catch (InterruptedException i) {
         }
         System.out.println(testPool.poolManager.toXml());
         if (testPool.poolManager.getNumBusy() != 0 || testPool.poolManager.getNumIdle() != 3)
            PoolManager.panic("TEST 1.3 FAILED: Wrong number of busy/idle resources");
         testPool.reserve();
         if (testPool.poolManager.getNumBusy() != 1 || testPool.poolManager.getNumIdle() != 2)
            PoolManager.panic("TEST 1.4 FAILED: Wrong number of busy/idle resources");
         try {
            Thread.currentThread().sleep(1000);
         }
         catch (InterruptedException i) {
         }
         System.out.println(testPool.poolManager.toXml());
      }


      // [2] Test with supplied instance ID and busy timeout
      if (true) {
         System.out.println("\n\n=========================\nStarting TEST 2 ...");
         TestPool testPool = new TestPool(3, 2000, 0);
         TestResource r0 = testPool.reserve("ID-0");
         TestResource r1 = testPool.reserve("ID-1");
         TestResource r2 = testPool.reserve("ID-2");
         r2 = testPool.reserve("ID-2");
         r2 = testPool.reserve("ID-2");
         r2 = testPool.reserve("ID-2");
         if (testPool.poolManager.getNumBusy() != 3 || testPool.poolManager.getNumIdle() != 0)
            PoolManager.panic("TEST 2.1 FAILED: Wrong number of busy/idle resources");
         testPool.reserve("ID-3");
         testPool.release(r0);
         if (testPool.poolManager.getNumBusy() != 2 || testPool.poolManager.getNumIdle() != 1)
            PoolManager.panic("TEST 2.2 FAILED: Wrong number of busy/idle resources");
         testPool.reserve("ID-4");
         testPool.release(r2);
         System.out.println(testPool.poolManager.toXml());



         // The resources are swapped to idle in 2 seconds, lets wait 3 seconds ...
         try {
            Thread.currentThread().sleep(3000);
         }
         catch (InterruptedException i) {
         }
         System.out.println(testPool.poolManager.toXml());
         if (testPool.poolManager.getNumBusy() != 0 || testPool.poolManager.getNumIdle() != 3)
            PoolManager.panic("TEST 2.3 FAILED: Wrong number of busy/idle resources");
         testPool.reserve("ID-5");
         if (testPool.poolManager.getNumBusy() != 1 || testPool.poolManager.getNumIdle() != 2)
            PoolManager.panic("TEST 2.4 FAILED: Wrong number of busy/idle resources");
         testPool.poolManager.destroy();
         System.out.println(testPool.poolManager.toXml());
         if (testPool.poolManager.getNumBusy() != 0 || testPool.poolManager.getNumIdle() != 0)
            PoolManager.panic("TEST 2.5 FAILED: Wrong number of busy/idle resources");
         testPool.reserve("ID-6");
         System.out.println(testPool.poolManager.toXml());
         if (testPool.poolManager.getNumBusy() != 1 || testPool.poolManager.getNumIdle() != 0)
            PoolManager.panic("TEST 2.6 FAILED: Wrong number of busy/idle resources");
         try {
            Thread.currentThread().sleep(1000);
         }
         catch (InterruptedException i) {
         }
         System.out.println(testPool.poolManager.toXml());
      }

      // [3] Test with supplied instance ID, busy timeout and erase timeout
      if (true) {
         System.out.println("\n\n=========================\nStarting TEST 3 ...");
         TestPool testPool = new TestPool(3, 2000, 3000); // erase resource after 3 sec in idle state
         TestResource r0 = testPool.reserve("ID-0");
         TestResource r1 = testPool.reserve("ID-1");
         TestResource r2 = testPool.reserve("ID-2");
         testPool.poolManager.erase("ID-2"); // erase from busy list
         System.out.println(testPool.poolManager.toXml());
         if (testPool.poolManager.getNumBusy() != 2 || testPool.poolManager.getNumIdle() != 0)
            PoolManager.panic("TEST 3.1 FAILED: Wrong number of busy/idle resources");

         // The resources are swapped to idle in 2 seconds, lets wait 3 seconds ...
         try {
            Thread.currentThread().sleep(3000);
         }
         catch (InterruptedException i) {
         }
         testPool.poolManager.erase("ID-1"); // erase from idle list
         System.out.println(testPool.poolManager.toXml());
         if (testPool.poolManager.getNumBusy() != 0 || testPool.poolManager.getNumIdle() != 1)
            PoolManager.panic("TEST 3.2 FAILED: Wrong number of busy/idle resources");

         // The resources are erased after 3 seconds in idle state, lets wait 4 seconds ...
         try {
            Thread.currentThread().sleep(4000);
         }
         catch (InterruptedException i) {
         }
         System.out.println(testPool.poolManager.toXml());
         if (testPool.poolManager.getNumBusy() != 0 || testPool.poolManager.getNumIdle() != 0)
            PoolManager.panic("TEST 3.3 FAILED: Wrong number of busy/idle resources");
         try {
            Thread.currentThread().sleep(1000);
         }
         catch (InterruptedException i) {
         }
         System.out.println(testPool.poolManager.toXml());
      }

      // [4] Test multi threaded
      if (true) {
         System.out.println("\n\n=========================\nStarting TEST 4 ...");
         class Test extends Thread {
            String ME = "TestThread";
            TestPool testPool;
            Test(String name, TestPool testPool) {
               this.ME = name;
               this.testPool = testPool;
            }
            public void run() {
               try {
                  System.out.println("\n\nStarted " + ME + " ...");
                  TestResource r0 = testPool.reserve(USE_OBJECT_REF);
                  TestResource r1 = testPool.reserve(USE_OBJECT_REF);
                  TestResource r2 = testPool.reserve(USE_OBJECT_REF);
                  TestResource r3 = testPool.reserve(USE_OBJECT_REF);
                  TestResource r4 = testPool.reserve(USE_OBJECT_REF);
                  org.jutils.runtime.Sleeper.sleep(20L);

                  // multi thread access
                  testPool.poolManager.isBusy("unknown");
                  testPool.poolManager.isBusy("TestResource-1");
                  testPool.poolManager.busyRefresh("TestResource-1");
                  testPool.poolManager.busyRefresh("TestResource-1");
                  testPool.poolManager.busyRefresh("TestResource-1");
                  testPool.poolManager.getNumBusy();
                  testPool.poolManager.getNumIdle();
                  testPool.poolManager.getState();
                  testPool.poolManager.toXml();
                  testPool.poolManager.erase("" + r2); // the instanceId
                  r2 = testPool.reserve();
                  testPool.release(r2);
                  r2 = testPool.reserve();
                  testPool.release(r2);
                  testPool.release(r3);
                  System.out.println("\n\nFinished " + ME + " ...");
                  //System.out.println(testPool.poolManager.toXml());
               }
               catch (Throwable e) {
                  e.printStackTrace();
                  PoolManager.panic("\n++++++++++++++++++++++++\nTEST 4.1 FAILED: " + e.toString());
               }
            }
         }
         int numThreads = 6;
         long busyToIdle = 5000L;
         long idleToErase = 4000L;
         TestPool testPool = new TestPool(60, busyToIdle, idleToErase);
         for (int jj = 0; jj < 2; jj++) {
            System.out.println("\n run # " + jj);
            for (int ii = 0; ii < numThreads; ii++) {
               String name = "TestThread-" + ii;
               Test p = new Test(name, testPool);
               p.setDaemon(true);
               p.start();
            }
            org.jutils.runtime.Sleeper.sleep(2000L);
            if (testPool.poolManager.getNumBusy() != 3 * numThreads)
               PoolManager.panic("TEST 4.2 FAILED: Wrong number " + testPool.poolManager.getNumBusy() + " of busy resources");
            if (jj == 0)
               org.jutils.runtime.Sleeper.sleep(4000L); // now all busy resources are idle
         }
         org.jutils.runtime.Sleeper.sleep(1000L);
         testPool.poolManager.destroy();
         System.out.println(testPool.poolManager.toXml());
         if (testPool.poolManager.getNumBusy() != 0 || testPool.poolManager.getNumIdle() != 0)
            PoolManager.panic("TEST 4.3 FAILED: Wrong number of busy/idle resources");
      }

      // [5] Test destroy
      if (true) {
         System.out.println("\n\n=========================\nStarting TEST 5 ...");
         TestPool testPool = new TestPool(10, 2000, 3000); // erase resource after 3 sec in idle state
         TestResource r0 = testPool.reserve(USE_HASH_CODE);
         TestResource r1 = testPool.reserve(GENERATE_RANDOM);
         TestResource r2 = testPool.reserve(USE_HASH_CODE);
         TestResource r3 = testPool.reserve(USE_OBJECT_REF);
         TestResource r4 = testPool.reserve();
         testPool.release(r2);
         r2 = testPool.reserve();
         testPool.release(r2);
         testPool.release(r3);
         System.out.println(testPool.poolManager.toXml());
         testPool.poolManager.destroy();
         System.out.println(testPool.poolManager.toXml());
         if (testPool.poolManager.getNumBusy() != 0 || testPool.poolManager.getNumIdle() != 0)
            PoolManager.panic("TEST 5.1 FAILED: Wrong number of busy/idle resources");
      }
   }

   private static void panic(String text) {
      System.out.println(text);
      System.exit(1);
   }
}
