package org.jutils.runtime;

/*------------------------------------------------------------------------------
Name:      Memory.java
Project:   jutils.org
Copyright: jutils.org, see jutils-LICENSE file
Comment:   Handling the Client data
Version:   $Id: Memory.java,v 1.4 2000/09/26 10:01:52 winkler Exp $
------------------------------------------------------------------------------*/
import java.io.*;


/**
 * Some helper methods for memory handling.
 * May be used for test clients etc.
 * @author Marcel Ruff
 */
public class Memory
{
   private final static String ME = "Memory";
   private long startMemory = 0L;



    /**
       * Create an instance only if you want to use it with the usedMemory() method.
       */
    public Memory() {
        startMemory = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
    }



    /**
      * Build a nice, human readable string for the size in MB/KB/Bytes.
      * <br><b>Example:</b><br>
      * <code>System.out.println(Memory.DataLenStr(136000));</code><br>
      *  -> "136 KB"
      * @param size is the size in bytes
      * @return a nice readable memory string
      */
    public static final String byteString(long size) {
        long gBytes = size / 1000000000L;
        long mBytes = size % 1000000000L / 1000000L;
        long kBytes = size % 1000000L / 1000L;
        long bytes = size % 1000L;
        String str;
        if (gBytes != 0)
            str = "" + gBytes + "." + Math.abs(mBytes) + " GBytes";
        else
            if (mBytes != 0)
                str = "" + mBytes + "." + Math.abs(kBytes) + " MBytes";
            else
                if (kBytes != 0)
                    str = "" + kBytes + "." + Math.abs(bytes) + " KBytes";
                else
                    str = "" + bytes + " Bytes";
        return str;
    }



    /**
      * Access a nice, human readable string with the current RAM memory situation.
      * @return a nice readable memory statistic string
      */
    public static final String getStatistic() {
        StringBuffer statistic = new StringBuffer();
        statistic.append("Total memory allocated = ");
        statistic.append(byteString(Runtime.getRuntime().totalMemory()));
        statistic.append(".");
        statistic.append(" Free memory available = ");
        statistic.append(byteString(Runtime.getRuntime().freeMemory()));
        statistic.append(".");
        return statistic.toString();
    }



    /**
       * Testsuite for Memory class.
       * <p />
       * Invoke: java org.jutils.util.Memory
       */
    public static void main(String[] args) {
        final String ME = "Memory";
        long size;
        size = 10L;
        System.out.println("Converting " + size + " to " + Memory.byteString(size));
        size = 999L;
        System.out.println("Converting " + size + " to " + Memory.byteString(size));
        size = 1999L;
        System.out.println("Converting " + size + " to " + Memory.byteString(size));
        size = -1999L;
        System.out.println("Converting " + size + " to " + Memory.byteString(size));
        size = 19998907L;
        System.out.println("Converting " + size + " to " + Memory.byteString(size));
        size = -779998907L;
        System.out.println("Converting " + size + " to " + Memory.byteString(size));
        size = 19998907000L;
        System.out.println("Converting " + size + " to " + Memory.byteString(size));
    }



    /**
       * Display the consumed memory (RAM) since construction of this object.
       * <p />
       * Note that this number is not very precise, since the garbage collector
       * is spitting in our soup.
       */
    public long usedMemory() {
        long endMemory = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
        return endMemory - startMemory;
    }



    /**
      * Lists the current RAM memory situation to System.out.
      */
    public static final void listStatistic() {
        listStatistic(new PrintWriter(System.out));
        return;
    }



    /**
      * Lists the current RAM memory situation to the passed PrintWriter.
      */
    public static final void listStatistic(PrintWriter out) {
        out.println("Total memory allocated = " + byteString(Runtime.getRuntime().totalMemory()));
        out.println("Free  memory available = " + byteString(Runtime.getRuntime().freeMemory()));
        return;
    }
}

