/*------------------------------------------------------------------------------
Name:     XmlComposition.java
Project:  doubleSlash Content Management
Copyright:This is proprietary source code of doubleSlash Net-Business GmbH
Copyright 2001 -  all right reserved
Autor:    Michael Winkler
          Oliver Belikan
Version:  $Revision: 1.10 $
Datum:    $Date: 2001/08/10 16:32:06 $
Comment:  Helperclass which helps to create an XML-Stream for use in
           servlets.
           Typical XML-Descripton for servlets:
           <page id=''>
              <data>                              // values which inserted by user, calculated ......
              </data>
              <settings>                          // textvalues from application
                 <label id=''....></label>        // just an simple text
                 <link id=''>                     // normal http-link
                    <url></url>                   // shown text from link
                    <label></label>
                 </link>
              </settings>
              <session>                           // userdata from the session
                 <language></language>
              </session>
              <description>                       // global settings like windowtitle, headline .....
              </description>
              <error>                             // errordescription if one occured
              </error>
           </page>
--------------------------------------------------------------------------------*/
package org.jutils.servlet;
import java.io.*;
import java.util.*;
import java.util.Enumeration;
import org.jutils.JUtilsException;
import org.jutils.time.DateTime;


/**
 * this is the class which you can create an nice xml output-stream.
 * Especially used in servlets to print out xml-stream to process
 * with xsl files
 */
public class XmlComposition {
   String id                     = "";
   StringBuffer data             = null;
   StringBuffer settings         = null;
   StringBuffer session          = null;
   StringBuffer error            = null;
   StringBuffer description      = null;

   private String language       = null;

   /**
    * XmlComposition
    */
   public XmlComposition() {
       data = new StringBuffer();
       settings = new StringBuffer();
       session = new StringBuffer();
       error = new StringBuffer();
       description = new StringBuffer();
   }

   /**
    * XmlComposition
    */
   public XmlComposition(String id) {
       this();
       this.id = id;
   }

   /**
    * addData
    */
   public void addData (String data) {
       this.data.append(data);
   }

   /**
    * addDescription
    */
   public void addDescription (String description) {
       this.description.append(description);
   }

   /**
    * addError
    */
   public void addError(String error) {
       this.error.append("   <error>"+ error + "</error>\n");
   }

   /**
    * addWarning
    */
   public void addWarning (String warning) {
      this.error.append("   <warning>"+ warning + "</warning>\n");
   }

   /**
    * addSession
    */
   public void addSession(String session) {
       this.session.append(session);
   }

   /**
    * appends param tag to the session. <param id="key">value</param>
    * Parameters are submitted form values.
    * @param key the name of the field
    * @param value the value of the http value
    */

    public void addParam (String key, String value) {
       this.session.append ("   <param id=\\"+key+"\">"+ value +"</param>\n");
    }

   /**
    * addSettings
    */
   public void addSettings(String settings) {
       this.settings.append(settings);
   }

   /**
    * adds a text field in the settings section
    * <text id ="">
    *    <teaser></teaser>
    *    <title></title>
    *    <summary><summary>
    *    <copytext></copytext>
    * </text>
    * @param textid id of the text
    * @param title title of text or news
    * @param teaser short description (one or two sentences) to show on overview pages
    * @param summary first part of text. Used in pages to read only first lines of text... summary and copytext
    * together is the whole text information
    * @param copytext the text it self (without summary)
    */
    public void addText (String textid, String title, String teaser, String summary, String copytext) {
       String str = "<text id=\""+ textid + ">";
       if (teaser != null && !teaser.equals(""))
          str = str + "   <teaser>"+teaser+"</teaser>\n";

       if (title != null && !title.equals(""))
          str = str + "   <title>"+title+"</title>\n";

       if (summary != null && !summary.equals(""))
          str = str + "   <summary>"+summary+"</summary>\n";

       if (copytext != null && !copytext.equals(""))
          str = str + "   <copytext>"+copytext+"</copytext>\n";

      str = str + "</text>\n";

       this.addSettings(str);
    }

    /**
     * adds a radiobutton to the settings
     * <radiobutton value=""></radiobutton>
     * @param name the name of the button
     * @param the value of the button
     */
     public void addRadioButton (String name, String value) {
        String str = "   <radiobutton value=\""+value+">"+name+"</radiobutton>\n";
        this.addSettings(str);
     }
    /**
     * adds a radiobutton to the settings
     * <radiobutton class="" value=""></radiobutton>
     * @param name the name of the button
     * @param value the value of the button
     * @param group radiobuttons with the same group viewed together.
     * (necessary if only one button will be selected)
     */
     public void addRadioButton (String name, String value, String group) {
        String str = "   <radiobutton class=\""+ group +"\" value=\""+value+">"+name+"</radiobutton>\n";
        this.addSettings(str);
     }
    /**
     * adds a radiobutton to the settings
     * <checkbox value=""></checkbox>
     * @param name the name of the button
     * @param the value of the button
     */
     public void addCheckBox (String name, String value) {
        String str = "   <checkbox value=\""+value+">"+name+"</checkbox>\n";
        this.addSettings(str);
     }
    /**
     * adds a radiobutton to the settings
     * <checkbox class="" value=""></checkbox>
     * @param name the name of the button
     * @param value the value of the button
     * @param group checkbox with the same group viewed together.
     * (necessary if only one button will be selected)
     */
     public void addCheckBox (String name, String value, String group) {
        String str = "   <checkbox class=\""+ group +"\" value=\""+value+">"+name+"</checkbox>\n";
        this.addSettings (str);
     }
    /**
     * adds a radiobutton to the settings
     * <select value=""></select>
     * @param name the name of the button
     * @param the value of the button
     */
     public void addSelect (String name, String value) {
        String str = "   <select value=\""+value+">"+name+"</select>\n";
        this.addSettings(str);
     }
    /**
     * adds a radiobutton to the settings
     * <select class="" value=""></select>
     * @param name the name of the selectbox
     * @param value the value of the select
     * @param group selects with the same group viewed together as html <option> elements.
     */
     public void addSelect (String name, String value, String group) {
        String str = "   <select class=\""+ group +"\" value=\""+value+">"+name+"</select>\n";
        this.addSettings (str);
     }
   /**
    * setLanguage
    */
   public void setLanguage (String lang) {
      this.language = lang;
   }

   /**
    * addLabel
    * adds an label to the section of settings
    * <settings>
    *    <label id="...">
    *    </label>
    * </settings>
    */
   public void addLabel(String id, String text)
   {
      String str = "";
      str += "\n   <label id =\""+ id + "\">";
      str += "\n      " + text;
      str += "\n   </label>\n";
      this.addSettings (str);
   }

   /**
    * Pictures given in xml-stream.
    * @param name of the picture (without extension gif or jpg)
    * @param path path to the picture on the server
    * @param type type of picture (gif or jpg...)
    */
   public void addImage(String name, String id, String path, String type) {
      String str ="";
      str += "   <img id=\""+id+"\" path=\""+path+"\" type=\""+type+"\">";
      str += "      " + name;
      str += "   </img>\n";
      this.addSession(str);
   }

   /**
    * adds a date information to the session
    * datevalue must be in Format dd.MM.yyyy
    */
    public void addDate (String datevalue) {
      try {
         String str = new DateTime (datevalue, "dd.MM.yyyy").getAsXml();
         this.addSession (str);
      }
      catch (JUtilsException e) {
         // do nothing but be aware, this occures if datevalue is not
         // in corrct value dd.MM.yyyy
      }
    }

   /**
    * addLabels
    * adds all labels found by the given substring in the resourcebundle
    * @param String pattern is an substring of the keys found in the given ResourceBundle
    * @param ResourceBundle rb the bundle in which the strings searched by the given pattern
    * @exception NullPointerException if <code>key</code> is
    * <code>null</
    * example:
    *    instead of three statements :
    *
    *    page.addLabel("prodrive.text.slg", languageText.getString("prodrive.text.slg"));
    *    page.addLabel("prodrive.text.velocity", languageText.getString("prodrive.text.velocity"));
    *    page.addLabel("prodrive.text.pressure", languageText.getString("prodrive.text.pressure"));
    *
    *    you could call :
    *
    *    page.addLabels ("prodrive.text", languageText);
    */
   public void addLabels (String pattern, ResourceBundle rb) {
      String str="";
      for (Enumeration e = rb.getKeys() ; e.hasMoreElements() ;) {
         str = (String)e.nextElement();
         if (str.regionMatches(true,0,pattern,0,pattern.length())) {
            // yes the pattern found in the string of the
            // resourcebundle
            this.addLabel (str, rb.getString (str));
         }
      }
   }
   /**
    * addLink
    * adds an link to the section of settings
    * <settings>
    *    <link id="...">
    *       <url>...</url>
    *       <text>...</text>
    *    </link>
    * </settings>
    */
   public void addLink (String id, String url, String text)
   {
      String str = "";
      str += "\n   <link id =\""+ id + "\">";
      str += "\n      <url>" + url + "</url>";
      str += "\n      <text>" + text + "</text>";
      str += "\n   </link>";
      this.addSettings (str);
   }
   /**
    * addLink
    * adds an link to the section of settings
    * <settings>
    *    <link id="...">
    *       <url>...</url>
    *       <text>...</text>
    *    </link>
    * </settings>
    */
   public void addLink (String id, String link) {
      String str = "";
      str += "\n   <link id =\""+ id + "\">";
      str += "\n      " + link;
      str += "\n   </link>";
      this.addSettings (str);
   }
   /**
    * adds simple link to the settings section
    *    <link id="..." url="...">
    *       <labe>...</label>
    *    </link>
    */
    public void addLinkLabel (String id, String url, String label) {
      String str = "";
      str += "   <link id =\""+ id + "\" url=\""+url+">\n";
      str += "      <label>" + label + "</label>\n";
      str += "   </link>\n";
      this.addSettings (str);
    }
   /**
    * adds link to the settings section with img
    *    <link id="..." url="...">
    *       <img path="..." type="...">...</img>
    *       <labe>...</label>
    *    </link>
    */
    public void addLinkLabel (String id, String url, String label, String imgpath, String imgtype, String imgname) {
      String str = "";
      str += "   <link id =\""+ id + "\" url=\""+url+">\n";
      str += "      <img path=\"" + imgpath + "\" type=\""+imgtype+"\">"+imgname+"<img>\n";
      str += "      <label>" + label + "</label>\n";
      str += "   </link>\n";
      this.addSettings (str);
    }

 /**
    * addLinks
    * adds all links found by the given substring in the resourcebundle
    * @param String pattern is an substring of the keys found in the given ResourceBundle
    * @param ResourceBundle rb the bundle in which the strings searched by the given pattern
    * @exception NullPointerException if <code>key</code> is
    * <code>null</
    * example:
    *    instead of three statements :
    *
    *    page.addLink ("prodrive.site.slg", languageText.getString("prodrive.site.slg.url", languageText.getString("prodrive.site.slg.text"));
    *    page.addLink ("prodrive.site.velocity", languageText.getString("prodrive.site.velocity.url", languageText.getString("prodrive.site.velocity.text"));
    *    page.addLink ("prodrive.site.pressure", languageText.getString("prodrive.site.pressure.url", languageText.getString("prodrive.site.pressure.text"));
    *
    *    you could call :
    *
    *    page.addLinks("prodrive.site", languageText);
    */
   public void addLinks (String pattern, ResourceBundle rb) {
      String str="", url="", text="";
      for (Enumeration e = rb.getKeys() ; e.hasMoreElements() ;) {
         str = (String)e.nextElement();
         if (str.regionMatches(true,0,pattern,0,pattern.length())) {
            // yes the pattern found in the string of the
            // resourcebundle
            url = rb.getString (str + ".url");
            text = rb.getString (str + ".text");
            this.addLink (str, url, text);
         }
      }
   }
   /**
    * clear
    */
   public void clear() {
       data = new StringBuffer();
       settings = new StringBuffer();
       session = new StringBuffer();
       description = new StringBuffer();
       error = new StringBuffer();
       this.language = null;
   }

   /**
    * toString
    */
   public String toString() {
       StringBuffer sb = new StringBuffer();
       StringBuffer tmp = new StringBuffer();
       sb.append("<?xml version='1.0' encoding='iso-8859-1' ?>\n");
       if (id == null)
           tmp.append("");
       else
           tmp.append(" id='").append(id).append("'");
       sb.append("<page").append(tmp).append(" xmlns:html='http://www.w3.org/TR/REC-html40/loose.dtd'>\n");
       sb.append("   <description>\n");
       if (settings != null) {
           sb.append(description);
       }
       sb.append("   </description>\n");
       sb.append("   <session>\n");
       if (session != null) {
           sb.append(session);
       }
       if (this.language != null) {
          sb.append ("<language>" + this.language + "</language>");
       }
       sb.append("   </session>\n");
       sb.append("   <data>\n");
       if (data != null) {
           sb.append(data);
       }
       sb.append("   </data>\n");
       sb.append("   <settings>\n");
       if (settings != null) {
           sb.append(settings);
       }
       sb.append("   </settings>\n");
       if (error != null) {
           sb.append("   <error>").append(error).append("</error>");
       }
       sb.append("</page>\n");
       return sb.toString();
   }

   /**
    *
   */
   public static void main(String args[]) throws Exception {
        XmlComposition p = new XmlComposition();
        p.addWarning("warning");
        p.addError("Error");
        p.addCheckBox("name", "value", "group");
        p.addParam("key1", "value1");
        p.addParam("key2", "value2");
        p.addSelect("name", "value");
        p.addRadioButton("name", "value", "group");
        System.out.println (p.toString());
   }
}

